/* -*-c-*-
 *
 * $Id: key-binary.c,v 1.5 2004/04/01 12:50:09 mdw Exp $
 *
 * Key binary encoding
 *
 * (c) 1999 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------* 
 *
 * $Log: key-binary.c,v $
 * Revision 1.5  2004/04/01 12:50:09  mdw
 * Add cyclic group abstraction, with test code.  Separate off exponentation
 * functions for better static linking.  Fix a buttload of bugs on the way.
 * Generally ensure that negative exponents do inversion correctly.  Add
 * table of standard prime-field subgroups.  (Binary field subgroups are
 * currently unimplemented but easy to add if anyone ever finds a good one.)
 *
 * Revision 1.4  2004/03/28 01:58:47  mdw
 * Generate, store and retreive elliptic curve keys.
 *
 * Revision 1.3  2001/02/03 11:57:00  mdw
 * Track mLib change: symbols no longer need to include a terminating
 * null.
 *
 * Revision 1.2  2000/06/17 11:25:20  mdw
 * Use secure memory interface from MP library.
 *
 * Revision 1.1  2000/02/12 18:21:02  mdw
 * Overhaul of key management (again).
 *
 */

/*----- Header files ------------------------------------------------------*/

#include <stdlib.h>
#include <string.h>

#include <mLib/bits.h>
#include <mLib/dstr.h>
#include <mLib/sub.h>
#include <mLib/sym.h>

#include "key-data.h"
#include "mp.h"
#include "mptext.h"

/*----- Main code ---------------------------------------------------------*/

/* --- @key_decode@ --- *
 *
 * Arguments:	@const void *p@ = pointer to buffer to read
 *		@size_t sz@ = size of the buffer
 *		@key_data *k@ = pointer to key data block to write to
 *
 * Returns:	Zero if everything worked, nonzero otherwise.
 *
 * Use:		Decodes a binary representation of a key.
 */

int key_decode(const void *p, size_t sz, key_data *k)
{
  const octet *q = p;
  size_t psz;
  unsigned e;

  /* --- Parse the header information --- *
   *
   * Make sure the size matches external reality.  Security holes have been
   * known to creep in without this sort of check.  (No, this isn't an after-
   * the-fact patch-up.)
   */

  e = LOAD16(q);
  psz = LOAD16(q + 2);
  if (psz + 4 > sz)
    return (-1);
  k->e = e;

  /* --- Now decide what to do --- */

  switch (e & KF_ENCMASK) {

    /* --- Plain binary data --- */

    case KENC_BINARY:
    case KENC_ENCRYPT:
      k->u.k.k = sub_alloc(psz);
      memcpy(k->u.k.k, q + 4, psz);
      k->u.k.sz = psz;
      break;

    /* --- Multiprecision integer data --- */

    case KENC_MP:
      k->u.m = mp_loadb(k->e & KF_BURN ? MP_NEWSEC : MP_NEW, q + 4, psz);
      break;

    /* --- String data --- */

    case KENC_STRING:
      k->u.p = xmalloc(sz + 1);
      memcpy(k->u.p, q + 4, sz);
      k->u.p[sz] = 0;
      break;

    /* --- Elliptic curve point data --- */

    case KENC_EC: {
      size_t xsz, ysz;
      EC_CREATE(&k->u.e);
      if (!sz) break;
      if (sz < 2) return (-1);
      xsz = LOAD16(q + 4);
      if (sz < xsz + 4) return (-1);
      ysz = LOAD16(q + 6 + xsz);
      if (sz < xsz + ysz + 4) return (-1);
      k->u.e.x = mp_loadb(MP_NEW, q + 6, xsz);
      k->u.e.y = mp_loadb(MP_NEW, q + 6 + xsz, ysz);
    } break;

    /* --- Structured key data --- */

    case KENC_STRUCT: {
      dstr d = DSTR_INIT;
      key_struct *ks;
      unsigned f;

      if ((k->e & ~KF_ENCMASK) || (psz & 3))
	return (-1);
      q += 4;
      sym_create(&k->u.s);

      while (psz) {

	/* --- Read the tag string --- */

	DRESET(&d);
	sz = LOAD8(q);
	if (sz >= psz)
	  goto fail;
	DPUTM(&d, q + 1, sz);
	DPUTZ(&d);
	sz = (sz + 4) & ~3;
	q += sz; psz -= sz;

	/* --- Read the encoding and size --- */

	e = LOAD16(q);
	sz = (LOAD16(q + 2) + 7) & ~3;
	if (sz > psz)
	  goto fail;

	/* --- Create a table node and fill it in --- */

	ks = sym_find(&k->u.s, d.buf, d.len, sizeof(*ks), &f);
	if (f)
	  goto fail;
	if (key_decode(q, sz, &ks->k)) {
	  sym_remove(&k->u.s, ks);
	  goto fail;
	}
	psz -= sz;
	q += sz;
      }
      dstr_destroy(&d);
      break;

      /* --- Tidy up after a failure --- */

    fail:
      dstr_destroy(&d);
      key_destroy(k);
      return (-1);
    } break;

    /* --- Everything else --- */

    default:
      return (-1);
  }

  /* --- OK, that was good --- */

  return (0);
}

/* --- @key_encode@ --- *
 *
 * Arguments:	@key_data *k@ = pointer to key data block
 *		@dstr *d@ = pointer to destination string
 *		@const key_filter *kf@ = pointer to key selection block
 *
 * Returns:	Nonzero if an item was actually written.
 *
 * Use:		Encodes a key block as binary data.
 */

int key_encode(key_data *k, dstr *d, const key_filter *kf)
{
  int rc = 0;
  if (!KEY_MATCH(k, kf))
    return (0);
  switch (k->e & KF_ENCMASK) {
    case KENC_BINARY:
    case KENC_ENCRYPT: {
      char *p;

      DENSURE(d, (k->u.k.sz + 7) & ~3);
      p = d->buf + d->len;
      STORE16(p, k->e);
      STORE16(p + 2, k->u.k.sz);
      d->len += 4;
      DPUTM(d, k->u.k.k, k->u.k.sz);
      rc = 1;
    } break;

    case KENC_MP: {
      char *p;
      size_t sz = mp_octets(k->u.m);

      DENSURE(d, (sz + 7) & ~3);
      p = d->buf + d->len;
      STORE16(p, k->e);
      STORE16(p + 2, sz);
      mp_storeb(k->u.m, p + 4, sz);
      d->len += sz + 4;
      rc = 1;
    } break;

    case KENC_STRING: {
      char *p;
      size_t sz = strlen(k->u.p);

      DENSURE(d, (sz + 7) & ~3);
      p = d->buf + d->len;
      STORE16(p, k->e);
      STORE16(p + 2, sz);
      memcpy(p + 4, k->u.p, sz);
      d->len += sz + 4;
      rc = 1;
    } break;

    case KENC_EC: {
      char *p;
      size_t xsz, ysz;
      size_t sz;

      if (EC_ATINF(&k->u.e))
	sz = 0;
      else {
	xsz = mp_octets(k->u.e.x);
	ysz = mp_octets(k->u.e.y);
	sz = xsz + ysz + 4;
      }
      DENSURE(d, (sz + 7) & ~3);
      p = d->buf + d->len;
      STORE16(p, k->e);
      STORE16(p + 2, sz);
      if (!EC_ATINF(&k->u.e)) {
	STORE16(p + 4, xsz);
	mp_storeb(k->u.e.x, p + 6, xsz);
	STORE16(p + 6 + xsz, ysz);
	mp_storeb(k->u.e.y, p + 8 + xsz, ysz);
      }
      d->len += sz + 4;
      rc = 1;
    } break;

    case KENC_STRUCT: {
      size_t n;
      char *p;
      key_struct *ks;
      sym_iter i;

      n = d->len;
      DENSURE(d, 4);
      p = d->buf + n;
      STORE16(p, k->e & KF_ENCMASK);
      d->len += 4;
      for (sym_mkiter(&i, &k->u.s); (ks = sym_next(&i)) != 0; ) {
	size_t o = d->len;
	DENSURE(d, 1);
	*(octet *)(d->buf + d->len++) = strlen(SYM_NAME(ks));
	DPUTS(d, SYM_NAME(ks));
	while (d->len & 3)
	  DPUTC(d, 0);
	if (key_encode(&ks->k, d, kf))
	  rc = 1;
	else
	  d->len = o;
      }
      if (!rc)
        d->len = n;
      else {
	p = d->buf + n + 2;
	n = d->len - n - 4;
	STORE16(p,  n);
      }
    } break;
  }
  while (d->len & 3)
    DPUTC(d, 0);
  return (rc);
}

/*----- That's all, folks -------------------------------------------------*/
